package org.tbasket.auth

import io.circe.*
import io.circe.generic.auto.*
import io.circe.parser.*
import io.circe.syntax.*
import io.getquill.*
import io.getquill.context.qzio.ZioJdbcContext
import io.getquill.context.sql.idiom.SqlIdiom
import org.tbasket.InternalBasketServerException
import org.tbasket.data.User
import pdi.jwt.algorithms.JwtAsymmetricAlgorithm
import pdi.jwt.{JwtClaim, JwtZIOJson}
import zio.*
import zio.http.*
import zio.http.api.HttpCodec.Method
import zio.http.model.Status.{InternalServerError, Ok}
import zio.http.model.Version.Http_1_1
import zio.http.model.{Headers, Method}

import java.io.ByteArrayInputStream
import java.security.PublicKey
import java.util.UUID
import javax.sql.DataSource
import scala.collection.immutable.HashMap

enum AuthentificatorError:
  case ExpiredToken
  case InvalidEmitterResponse

case class JwtContent(uuid: UUID)

class Authentificator(url: URL, key: PublicKey, algorithm: JwtAsymmetricAlgorithm) {
  
  private def defineCustomClaims(user: User): String = {
    JwtContent(user.id).asJson.noSpaces.toString
  }
  
  private def mkRequest(user: User): Request = {
    val custom = defineCustomClaims(user)
    Request(Body.fromString(custom), Headers.empty, Method.GET, url, Http_1_1, None)
  }
  
  def requestJwt(user: User) = {
    Client.request(mkRequest(user))
      .flatMap {
        case Response(Ok, _, body, _, _)               =>
          body.asString
        case Response(InternalServerError, _, _, _, _) =>
          ZIO.fail(new InternalBasketServerException("internal server error"))
        case _                                         =>
          ZIO.fail(new InternalBasketServerException("Received unknown response from emitter"))
      }
  }
  
  def validateAndGetUser(jwt: String) = {
    for
    //decoding token
      claims <- ZIO.fromTry(JwtZIOJson.decode(jwt, key, Seq(algorithm)))
      //ensure that the token is not expired (or else fail)
      _ <- ZIO.attempt(claims.expiration)
        .someOrFail("Received invalid jwt token (missing expiration date)")
        .filterOrFail(_ <= java.lang.System.currentTimeMillis())("Expired token")
      //
      uuid <- ZIO.attempt(claims.content)
        .mapAttempt(decode[JwtContent](_))
        .flatMap(ZIO.fromEither(_))
        .map(_.uuid)
      
      user <- ZIO.serviceWithZIO[ZioJdbcContext[SqlIdiom, NamingStrategy]] { ctx =>
        import ctx._
        run(quote {
          query[User].filter(_.id == lift(uuid))
        }).map(_.headOption)
          .someOrFail("uuid not found.")
      }
    yield user
  }
}
