package org.tbasket.handler

import io.getquill.*
import io.getquill.context.ZioJdbc.*
import io.getquill.context.qzio.{ZioContext, ZioJdbcContext}
import io.getquill.context.sql.idiom.SqlIdiom
import org.tbasket.auth.Authentificator
import org.tbasket.handler.HandlerUtils.errorBody
import org.tbasket.handler.LoginError.*
import org.tbasket.data.User
import zio.http.*
import zio.http.model.{Cookie, Header, Headers, Status}
import zio.json.*
import zio.json.ast.Json.Str
import zio.json.ast.{Json, JsonCursor}
import zio.{ZEnvironment, ZIO, *}

import java.sql.SQLException
import java.util.UUID

enum LoginError:
  case TokenNotFound(token: UUID)
  case UserNotFound(user: User)
  case InvalidPassword
  case InvalidRequest(msg: String, cause: String)
  case InternalError(t: Throwable)


object LoginHandler:
  
  private def getUser(json: Json) =
    ZIO.serviceWithZIO[ZioJdbcContext[SqlIdiom, NamingStrategy]] { ctx =>
      import ctx.*
      for
        mail <-
          ZIO.fromEither(json.get[Str](JsonCursor.field("mail").isString).map(_.value))
            .mapError(InvalidRequest("Missing or invalid field mail", _))
        password <-
          ZIO.fromEither(json.get[Str](JsonCursor.field("password").isString).map(_.value.hashCode))
            .mapError(InvalidRequest("Missing or invalid field password", _))
        
        result <- run(quote { //              TODO use argon2id
          query[User]
            .filter(usr => usr.mailAddress == lift(mail))
            .filter(usr => usr.passwordHash == lift(password))
        }).mapError(InternalError.apply)
      yield result.headOption
    }.someOrFail(InvalidPassword)
  
  def post(request: Request) =
    val bindSession =
      for
        body <- request
          .body
          .asString
          .tapError(Console.printError(_))
          .mapError(s =>
            InvalidRequest("Wrong request body", s.getMessage)
          )
        
        json <- ZIO.fromEither(body.fromJson[Json])
          .mapError(InvalidRequest("Invalid JSON body", _))
        
        user <- getUser(json)
        jwt <- ZIO.serviceWithZIO[Authentificator](_.requestJwt(user))
      yield (user, jwt)
    
    bindSession.map { case (_, jwt) =>
      Response(
        status = Status.Found,
        headers = Headers.location("/") ++ //login successful, go back to main page
          Headers.setCookie(Cookie("JWT", jwt)) //and set the token cookie
      )
    } fold( {
      _ match
        case TokenNotFound(_)           => Response(
          status = Status.Unauthorized,
          body = errorBody("unauthorized", "unknown token"),
          headers =
            Headers(
              Headers.location("/login")
            ) // send back caller to login panel
        )
        case UserNotFound(_)            => Response(
          status = Status.Unauthorized,
          body = errorBody("unauthorized", "unknown user email"),
          headers =
            Headers(
              Headers.location("/register")
            ) // send back caller to register panel
        )
        case InvalidPassword            => Response(
          status = Status.Unauthorized,
          body = errorBody("unauthorized", "invalid password")
        )
        case InvalidRequest(msg, cause) => Response(
          status = Status.Unauthorized,
          body = errorBody("wrong request", s"$cause: $msg")
        )
        case InternalError(_)           => Response(
          status = Status.InternalServerError,
          body = errorBody("internal", "internal error, please contact support")
        )
    }, x => x)
